package org.nrg.xnat.restlet.extensions;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.nio.charset.Charset;
import java.sql.SQLException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.apache.commons.io.FileUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.nrg.xdat.base.BaseElement;
import org.nrg.xdat.model.CatCatalogI;
import org.nrg.xdat.model.CatEntryI;
import org.nrg.xdat.model.XnatAbstractresourceI;
import org.nrg.xdat.om.XnatProjectdata;
import org.nrg.xdat.om.XnatResourcecatalog;
import org.nrg.xdat.schema.SchemaElement;
import org.nrg.xft.ItemI;
import org.nrg.xft.XFT;
import org.nrg.xft.XFTItem;
import org.nrg.xft.XFTTable;
import org.nrg.xft.collections.ItemCollection;
import org.nrg.xft.db.MaterializedView;
import org.nrg.xft.event.EventDetails;
import org.nrg.xft.event.EventMetaI;
import org.nrg.xft.event.EventUtils;
import org.nrg.xft.event.persist.PersistentWorkflowI;
import org.nrg.xft.event.persist.PersistentWorkflowUtils;
import org.nrg.xft.exception.DBPoolException;
import org.nrg.xft.exception.ElementNotFoundException;
import org.nrg.xft.exception.FieldNotFoundException;
import org.nrg.xft.exception.InvalidValueException;
import org.nrg.xft.exception.XFTInitException;
import org.nrg.xft.schema.Wrappers.GenericWrapper.GenericWrapperElement;
import org.nrg.xft.schema.Wrappers.GenericWrapper.GenericWrapperField;
import org.nrg.xft.search.CriteriaCollection;
import org.nrg.xft.search.ItemSearch;
import org.nrg.xft.utils.DateUtils;
import org.nrg.xft.utils.FieldMapping;
import org.nrg.xft.utils.NodeUtils;
import org.nrg.xft.utils.SaveItemHelper;
import org.nrg.xft.utils.StringUtils;
import org.nrg.xft.utils.XMLUtils;
import org.nrg.xnat.helpers.resource.direct.DirectProjResourceImpl;
import org.nrg.xnat.restlet.XnatRestlet;
import org.nrg.xnat.restlet.resources.SecureResource;
import org.nrg.xnat.utils.CatalogUtils;
import org.restlet.Context;
import org.restlet.data.MediaType;
import org.restlet.data.Request;
import org.restlet.data.Response;
import org.restlet.data.Status;
import org.restlet.resource.StringRepresentation;
import org.restlet.resource.Variant;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import com.google.common.base.Charsets;


/*
 * Example Call:
 * http://nrg123:8080/xnat/REST/services/csvUploadFromCatalog?project=RECON&source-resourceId=6&source-path=mrDataSource.txt&xform-resourceId=6&xform-path=mrData2.txt&fieldMapping-resourceId=6&fieldMapping-path=mrFields.xml
 */


@XnatRestlet({"/services/csvUploadFromCatalog"})
public class CsvUploadFromCatalog extends SecureResource {
	
	private static final Log _log = LogFactory.getLog(CsvUploadFromCatalog.class);


	
	public final static String[] RESULT_HEADERS = {"ID","RESULT"};
	
	//the rest of these are all querystring parameters
	XnatProjectdata project=null;
	
	String xsiType = null;
	boolean ignore_checksums = false;
	boolean do_not_save = false;

	String source_resourceId = null;
	String source_path = null;
	String source_checksum = null;
	
	String xform_resourceId = null;
	String xform_path = null;
	String xform_checksum = null;
	
	String fieldMapping_resourceId = null;
	String fieldMapping_path = null;
	String fieldMapping_checksum = null;
		
	
	public CsvUploadFromCatalog(Context context, Request request, Response response) {
		
		super(context, request, response);
		
		String projectId = getQueryVariable("project");
		if(projectId!=null){
			project = XnatProjectdata.getProjectByIDorAlias(projectId, user, false);
		}
		if (project == null) {
			response.setStatus(Status.CLIENT_ERROR_NOT_FOUND);
			this.getResponse().setEntity(new StringRepresentation( "Unable to identify project " + projectId, MediaType.APPLICATION_JSON));
			return;
		}
		
		ignore_checksums = Boolean.parseBoolean(getQueryVariable("ignoreChecksums")); //not required. default is false
		xsiType = getQueryVariable("dataType");  //required if fieldMapping-resourceId is not set.
		do_not_save = Boolean.parseBoolean(getQueryVariable("doNotSave")); //not required. default is false

		//source isn't required
		source_resourceId = getQueryVariable("sourceResourceId");
		source_path = getQueryVariable("sourcePath");
		source_checksum = getQueryVariable("sourceChecksum");
		
		//xform's resourceId and path are required
		xform_resourceId = getQueryVariable("xformResourceId");
		xform_path = getQueryVariable("xformPath");
		xform_checksum = getQueryVariable("xformChecksum");
		
		//not necessary
		fieldMapping_resourceId = getQueryVariable("fieldMappingResourceId");
		fieldMapping_path = getQueryVariable("fieldMappingPath");
		fieldMapping_checksum = getQueryVariable("fieldMappingChecksum");

		this.getVariants().add(new Variant(MediaType.TEXT_HTML));
		this.getVariants().add(new Variant(MediaType.TEXT_XML));
		this.getVariants().add(new Variant(MediaType.APPLICATION_JSON));
	}
	
	//utility method to convert a file to an array of arrays of data values.
	//on error it adds an entry to the logs and returns null
	private ArrayList<ArrayList<String>> csvFileToArray(File csv, Charset cs){
	    ArrayList<ArrayList<String>> ret = new ArrayList<ArrayList<String>>();
	    BufferedReader br = null;
	    try{	
	    br = new BufferedReader(new FileReader(csv));
	    String r = null;
	    while(( r = br.readLine()) != null){
	    	if(!"".equals(r)){
	    		ret.add(StringUtils.CommaDelimitedStringToArrayList(r));
	    	}
	    }
	    } catch (FileNotFoundException ex){
	    	_log.error("csvFileToArray unable to find file",ex);
	    	
	    } catch (IOException ioe){
	    	_log.error("csvFileToArray io error",ioe);
	    } finally {
	    	try {
	    		if (br != null){
	    			br.close();
	    		}
	    		return ret;
	    	} catch (IOException ioe){
	    		_log.error("csvFileToArray cannot close file - FATAL ERROR",ioe);
	    	}
	    	
	    }
	    return null;  //we failed
	}
	
	//awesome, since this is a module, I can't just add a constructor to FieldMapping.
	//I have to do it manually here in what is almost a copy/paste from FieldMapping.populateFromFile
	private FieldMapping buildFieldMapping(String s){
		FieldMapping fm = new FieldMapping();
		Document dom =XMLUtils.GetDOM(s);
		Element e= dom.getDocumentElement();
		fm.setElementName( NodeUtils.GetAttributeValue(e, "data-type", "") );
		fm.setTitle( NodeUtils.GetAttributeValue(e, "title", "") );
		fm.setID( NodeUtils.GetAttributeValue(e, "ID", ""));
		String dateS= NodeUtils.GetAttributeValue(e, "create-date", "");
		try {
			fm.setCreateDate( DateUtils.parseDate(dateS) );
		} catch (ParseException e1) {
			_log.error("Unable to parse date in FieldMapping file.", e1);
		}
		ArrayList<String> fields = new ArrayList<String>();
		ArrayList<Node> nodes =NodeUtils.GetLevelNNodes(e, "field", 1);
		for(Node n : nodes){
			fields.add(NodeUtils.GetNodeText(n));
		}
		fm.setFields(fields);
		return fm;
	}
	
	//Utility class to help me handle multiple return values from getCatalogEntry
	public class Pair<T, U> {         
	    public final T t;
	    public final U u;

	    public Pair(T t, U u) {         
	        this.t= t;
	        this.u= u;
	     }
	 }
	
	/**
     * Method for retrieving a file and a catalog entry from a project based on a specific resource label and path
     * Returns a CatalogEntry/File Pair or throws an exception on failure.
     */
    public Pair<CatEntryI, File> getCatalogEntry( final String resource_label, final String relativePath) 
    		throws IOException, Exception{
    	
        final DirectProjResourceImpl impl= new DirectProjResourceImpl(project, false, user, (EventMetaI)null);

        XnatAbstractresourceI abst = null;
        
        //first, try to pull the catalog by ID. if that doesn't work, try to pull it by the label. I'm doing this
        //because listing the files for a resource give a URI with the ID instead of label. people may use the ID.
        try {
        	int id = Integer.parseInt(resource_label);
        	abst=impl.getResourceById(id, null);
        } catch (Exception e){
        	//we're ok, this means we couldn't parse the resource_label into an ID. next try it as a label
        }
        if(abst == null ){
        	abst=impl.getResourceByLabel(resource_label, null);
        }
        
        if(!(abst instanceof XnatResourcecatalog)){
            throw new Exception("getCatalogEntry - Unsupported resource format");
        }
        
        final File catalogFile=CatalogUtils.getCatalogFile(project.getArchiveRootPath(), (XnatResourcecatalog) abst);//retrieve the actual catalog file
        if(catalogFile==null){
            throw new IOException("getCatalogEntry - CatalogFile not found");
        }
        
        final CatCatalogI catalogBean=CatalogUtils.getCatalog(catalogFile);//parse the file into a bean
                
        final CatEntryI entry=CatalogUtils.getEntryByURI(catalogBean,relativePath);//retrieve the entry which matches the relative path
        if(entry==null){
            throw new IOException("getCatalogEntry - Entry not found");
        }
        return new Pair<CatEntryI, File>(entry, CatalogUtils.getFile(entry, catalogFile.getParentFile().getAbsolutePath()));  
    }

    private Pair<CatEntryI, File> getThatResource(String resourceId, String path, XFTTable results, String nameForLogging){
    	Pair<CatEntryI, File> thePair = null;
    	
		//grab the source file if it exists
		try{
			thePair = getCatalogEntry(resourceId, path); 			
		} catch(Exception e){
			//don't fail, but make sure you let the user know that the document doesn't exist.
			String[] tmp = new String[2];
        	tmp[0] = "WARNING";
        	tmp[1] = nameForLogging + " Document does not exist: " + e.getMessage();
        	results.rows().add(tmp);
		}
		return thePair;
    }
    
    private boolean isChecksumOk(Pair<CatEntryI, File> thePair, String expectedChecksum, String nameForLogging){
    	if(!ignore_checksums && !org.apache.commons.lang.StringUtils.equalsIgnoreCase(expectedChecksum,  thePair.t.getDigest())){
			//Fail on unequal checksums.
			_log.error(nameForLogging + " checksum does not match that which is stored in the resource. If you wish to continue, resubmit without a checksum.");
			this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
			this.getResponse().setEntity(new StringRepresentation( nameForLogging + " checksum does not match that which is stored in the resource. Fix or resubmit with igore-checksums=true" , MediaType.APPLICATION_JSON));
			return false;
		}
    	return true;
    }
    
	@Override
	public boolean allowPost() {
		return true;
	}
	
	@Override
	public void handlePost(){
		
		/*
		 * Here's what is going on.
		 * 1. grab everything you need to do your work. Fail if you have to.
		 * 2. convert the file you're going to import into an array of arrays. if that fails, bail.
		 * 3. figure out which column maps to which data. do this by either using the fieldMapping file or building your own off the data's header.
		 * 4. import the data using a method that is almost exactly like the import routine in CsvUpload2 (the OG spreadsheet uploader)
		 * 5. return all success/error messages to the user via the XFTTable called results you've been filling in all along.
		 */
		

		Pair<CatEntryI, File> xformPair = null;   //cat entry and file for the transformed csv. this is what we're going to import so it is not optional
		Pair<CatEntryI, File> sourcePair = null;  //catalog entry and file for the source document. Totally optional.
		Pair<CatEntryI, File> fieldMappingPair = null;  //cat entry and file for the field mapping file. Totally optional.
		
		//store all error messages and whatnot in an XFT Table that we will return to the caller.
		XFTTable results = new XFTTable();
        String[] headers=RESULT_HEADERS.clone();
        results.initTable(headers);
        
        xformPair = getThatResource(xform_resourceId, xform_path, results, "Transfomed");
        if(xformPair == null){
        	//Fail because the transformed file is absolutely required.
			this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
			this.getResponse().setEntity(new StringRepresentation( "Unable to read transformed file." , MediaType.APPLICATION_JSON));
			return;
        }
        if(! isChecksumOk(xformPair, xform_checksum, "Transformed")){
        	//Fail on unequal checksums.
			_log.error("Transformed checksum does not match that which is stored in the resource. If you wish to continue, resubmit without a checksum.");
			this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
			this.getResponse().setEntity(new StringRepresentation( "Transformed checksum does not match that which is stored in the resource. Fix or resubmit with igore-checksums=true" , MediaType.APPLICATION_JSON));
			return;
        }
        
        sourcePair = getThatResource(source_resourceId,source_path, results, "Source");
        
        //I know this isChecksumOk nonsense looks a lot like a ton of repeated code, BUT, if a checksum fails, I need to know which one so I can let the user know.

        if(! isChecksumOk(sourcePair, source_checksum, "Source")){
        	//Fail on unequal checksums.
			_log.error("Transformed checksum does not match that which is stored in the resource. If you wish to continue, resubmit without a checksum.");
			this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
			this.getResponse().setEntity(new StringRepresentation( "Transformed checksum does not match that which is stored in the resource. Fix or resubmit with igore-checksums=true" , MediaType.APPLICATION_JSON));
			return;
        }
        
		fieldMappingPair = getThatResource(fieldMapping_resourceId, fieldMapping_path, results, "Field Mapping");
        if(! isChecksumOk(fieldMappingPair, fieldMapping_checksum, "Field Mapping")){
        	//Fail on unequal checksums.
			_log.error("Field Mapping checksum does not match that which is stored in the resource. If you wish to continue, resubmit without a checksum.");
			this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
			this.getResponse().setEntity(new StringRepresentation( "Field Mapping checksum does not match that which is stored in the resource. Fix or resubmit with igore-checksums=true" , MediaType.APPLICATION_JSON));
			return;
        }
        
		//load the transformed CSV file as an array of arrays. (row/column data)
	    ArrayList<ArrayList<String>> data = csvFileToArray(xformPair.u, Charsets.UTF_8);
	    
	    if(data == null){
	    	//fail if you have problem with the transformed data file.
	    	_log.error("Unable to read transformed data file.");
			this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
			this.getResponse().setEntity(new StringRepresentation( "Unable to read transformed file." , MediaType.APPLICATION_JSON));
			return;
	    }
	    
	    //Either pull the field mapping file from a catalog or try to build one using the data's header
	    FieldMapping fm = null;
	    if(fieldMappingPair == null){
	    	//build a field mapping from the CSV's header
	    	String[] tmp = new String[2];
        	tmp[0] = "INFO";
        	tmp[1] = "Building FieldMapping file from data header.";
        	results.rows().add(tmp);
	    	
		    ArrayList<String> header = data.get(0);
			if (xsiType == null) {
				this.getResponse().setStatus(Status.CLIENT_ERROR_NOT_FOUND, "You must supply an data-type parameter OR a field mapping resource.");
				return;
			}
	    	fm = new FieldMapping();
		    fm.setElementName(xsiType);
		    fm.setFields(header);
		    fm.setTitle("TemporaryMappingFor"+xsiType);
		    fm.setID(UUID.randomUUID().toString());
		} else {
			//this builds a field mapping from the xml files stored as a catalog file
			try {
				fm = buildFieldMapping(FileUtils.readFileToString( fieldMappingPair.u, Charsets.UTF_8.displayName()));
			} catch (Exception e){
				_log.error("Unable to read field mapping file.", e);
				this.getResponse().setStatus(Status.CLIENT_ERROR_BAD_REQUEST);
				this.getResponse().setEntity(new StringRepresentation( "Unable to read field mapping file." , MediaType.APPLICATION_JSON));
				return;
			}
	    }
	    
	    //remove the header from the CSV file. We got what we wanted.
	    data.remove(0);
		
	    //import the data
	    almostCopyPasteFromCsvUpload2(data, fm, results);
	
		Hashtable<String,Object> params=new Hashtable<String,Object>();
        params.put("title", "Results");

        Map<String,Map<String,String>> cp = new Hashtable<String,Map<String,String>>();
        cp.put("URI", new Hashtable<String,String>());
        cp.get("URI").put("serverRoot", getContextPath());
        MediaType mt = overrideVariant(this.getVariants().get(0)); //TODO: I have no idea what I'm doing, here.
        this.getResponse().setEntity(this.representTable(results, mt, params,cp));
		this.getResponse().setStatus(Status.SUCCESS_OK);
		return;
	}
	
	//This method creates XFTItems using the data (rows) and field mapping. Results are saved to the XFTTable. 
	//as it says, this is very similar to the method in CsvUpload2 that does the same thing.
	//I copy-pasted because this is a module and I needed some additional functionality (like, reporting errors to a table instead of HTML)
	private void almostCopyPasteFromCsvUpload2(ArrayList<ArrayList<String>> rows, FieldMapping fm, XFTTable results) 
	{
        List fields = fm.getFields();
        try {
            String rootElementName = fm.getElementName();
            GenericWrapperElement.GetElement(rootElementName);
            
            Iterator iter = rows.iterator();
            while(iter.hasNext())
            {

                ArrayList row = (ArrayList)iter.next();
                ArrayList rowSummary = new ArrayList();
                XFTItem item = XFTItem.NewItem(rootElementName, user);
                Iterator iter2 = row.iterator();
                int columnIndex = 0;
                while (iter2.hasNext())
                {
                    String column = (String)iter2.next();
                    String xmlPath = (String)fields.get(columnIndex);
                    
                    if (!column.equals("")){
                        rowSummary.add(column);

                        GenericWrapperField gwf = null;
                        try {
                            gwf = GenericWrapperElement.GetFieldForXMLPath(xmlPath);
                        } catch (FieldNotFoundException e) {
                        }


                        if (gwf!=null && gwf.getBaseElement()!=null && !gwf.getBaseElement().equals("")){
                            try {
                                ItemSearch search = ItemSearch.GetItemSearch(gwf.getBaseElement(), user);
                                SchemaElement se =SchemaElement.GetElement(gwf.getBaseElement());
                                if ((project!=null && !project.getId().equals("")) && se.hasField(se.getFullXMLName() + "/sharing/share/project")){
                                    CriteriaCollection cc = new CriteriaCollection("OR");
                                    cc.addClause(se.getFullXMLName() + "/" + gwf.getBaseCol(), column);

                                    CriteriaCollection sub = new CriteriaCollection("AND");
                                    sub.addClause(se.getFullXMLName() + "/sharing/share/project", project.getId());
                                    sub.addClause(se.getFullXMLName() + "/sharing/share/label", column);
                                    cc.add(sub);
                                    
                                    sub = new CriteriaCollection("AND");
                                    sub.addClause(se.getFullXMLName() + "/project", project.getId());
                                    sub.addClause(se.getFullXMLName() + "/label", column);
                                    cc.add(sub);                                    

                                    search.add(cc);
                                }else{
                                    search.addCriteria(se.getFullXMLName() + "/" + gwf.getBaseCol(), column);
                                }

                                ItemCollection items =search.exec(false);

                                if (items.size()==1){
                                    item.setProperty(xmlPath,items.getFirst().getProperty("ID"));
                                    columnIndex++;
                                    continue;
                                }
                            } catch (Exception e) {
                            	String[] tmp = new String[2];
                            	tmp[0] = "ERROR";
                            	tmp[1] = "General Exception: " + e.getMessage();
                            	results.rows().add(tmp);
                            }

                        }

                        try {
                            item.setProperty(xmlPath, column);
                        } catch (FieldNotFoundException e) {
                            _log.error("",e);
                            String[] tmp = new String[2];
                        	tmp[0] = "ERROR";
                        	tmp[1] = "FieldNotFoundException: " + e.getMessage();
                        	results.rows().add(tmp);
                        } catch (InvalidValueException e) {
                            _log.error("",e);
                            String[] tmp = new String[2];
                        	tmp[0] = "ERROR";
                        	tmp[1] = "InvalidValueException: " + e.getMessage();
                        	results.rows().add(tmp);
                        }
                    }
                    columnIndex++;
                }

                if (project!=null && !project.getId().equals("")){
                    SchemaElement se = SchemaElement.GetElement(rootElementName);

                    if (se.hasField(rootElementName +"/sharing/share/project") && se.hasField(rootElementName +"/sharing/share/label")){
                        try {
                            String id = item.getStringProperty("ID");
                            if(item.getStringProperty("project")==null){
                                item.setProperty(rootElementName + "/project", project.getId());
                                if(item.getStringProperty("label")==null){
                                    item.setProperty(rootElementName + "/label", id);   
                                }                         	
                            }else{
                            	if(item.getStringProperty("project").equals(project.getId())){
                                    if(item.getStringProperty("label")==null){
                                        item.setProperty(rootElementName + "/label", id);   
                                    }                         	
                            	}else{
                            		item.setProperty(rootElementName + "/sharing/share/project", project.getId()); 
                            		item.setProperty(rootElementName + "/sharing/share/label", id); 
                            	}
                            }

                            ItemSearch search = ItemSearch.GetItemSearch(rootElementName, user);
                            CriteriaCollection cc = new CriteriaCollection("OR");
                            cc.addClause(se.getFullXMLName() + "/ID", id);


                            CriteriaCollection sub = new CriteriaCollection("AND");
                            sub.addClause(se.getFullXMLName() + "/sharing/share/project", project.getId());
                            sub.addClause(se.getFullXMLName() + "/sharing/share/label", id);
                            cc.add(sub);
                            
                            sub = new CriteriaCollection("AND");
                            sub.addClause(se.getFullXMLName() + "/project", project.getId());
                            sub.addClause(se.getFullXMLName() + "/label", id);
                            cc.add(sub);

                            search.add(cc);
                            ItemCollection items =search.exec(false);

                            if (items.size()>0){
                                item.setProperty("ID",items.getFirst().getProperty("ID"));
                            }else{
                            	if(item.getStringProperty("label")!=null && item.getStringProperty("label").equals(item.getStringProperty("ID"))){
                            		ItemI om = BaseElement.GetGeneratedItem(item);
                            		Class c = om.getClass();
                            		Object[] intArgs = new Object[] {};
                        			Class[] intArgsClass = new Class[] {};
                        			
                        			String newID=null;
                            		try {
                                        Method m = c.getMethod("CreateNewID",intArgsClass);
                                        if(m!=null){
                                            try {
                                                try {
                                                	newID =(String)m.invoke(null,intArgs);
                                                } catch (RuntimeException e3) {
                                                    _log.error("",e3);
                                                    String[] tmp = new String[2];
                                                	tmp[0] = "ERROR";
                                                	tmp[1] = "RuntimeException: " + e3.getMessage();
                                                	results.rows().add(tmp);
                                                }
                                            } catch (IllegalArgumentException e2) {
                                                _log.error("",e2);
                                                String[] tmp = new String[2];
                                            	tmp[0] = "ERROR";
                                            	tmp[1] = "IllegalArgumentException: " + e2.getMessage();
                                            	results.rows().add(tmp);
                                            } catch (InvocationTargetException e2) {
                                                _log.error("",e2);
                                                String[] tmp = new String[2];
                                            	tmp[0] = "ERROR";
                                            	tmp[1] = "InvocationTargetException: " + e2.getMessage();
                                            	results.rows().add(tmp);
                                            }
                                        }
                                    } catch (SecurityException e1) {
                                        _log.error("",e1);
                                        String[] tmp = new String[2];
                                    	tmp[0] = "ERROR";
                                    	tmp[1] = "SecurityException: " + e1.getMessage();
                                    	results.rows().add(tmp);
                                    } catch (NoSuchMethodException e1) {
                                        _log.error("",e1);
                                        String[] tmp = new String[2];
                                    	tmp[0] = "ERROR";
                                    	tmp[1] = "NoSuchMethodException: " + e1.getMessage();
                                    	results.rows().add(tmp);
                                    }
                                    
                                    if(newID!=null){
                                    	item.setProperty("ID", newID);
                                    }else{
                                    	item.setProperty("ID",XFT.CreateIDFromBase(XFT.GetSiteID(),5, "ID", se.getSQLName(), null,null));
                                    }
                            	}
                            }
                        } catch (FieldNotFoundException e) {
                            _log.error("",e);
                            String[] tmp = new String[2];
                        	tmp[0] = "ERROR";
                        	tmp[1] = "FieldNotFoundException: " + e.getMessage();
                        	results.rows().add(tmp);
                        } catch (InvalidValueException e) {
                            _log.error("",e);
                            String[] tmp = new String[2];
                        	tmp[0] = "ERROR";
                        	tmp[1] = "InvalidValueException: " + e.getMessage();
                        	results.rows().add(tmp);
                        } catch (Exception e) {
                            _log.error("",e);
                            String[] tmp = new String[2];
                        	tmp[0] = "ERROR";
                        	tmp[1] = "General Exception: " + e.getMessage();
                        	results.rows().add(tmp);
                        }
                    }
                }
                
                EventDetails ed = this.newEventInstance(EventUtils.CATEGORY.DATA, "Spreadsheet Upload");
                ed.setComment(String.format("source: %s %s  %s | xform: %s %s %s | fieldMapping: %s %s %s | ignore-checksums: %s", source_resourceId, source_path, source_checksum, xform_resourceId, xform_path, xform_checksum, fieldMapping_resourceId, fieldMapping_path, fieldMapping_checksum, ignore_checksums));
                
                
                PersistentWorkflowI wrk= null;
                try {
                	
                	if(do_not_save){
                		String[] tmp = new String[2];
	                    tmp[0] = "SUCCESS";
	                    tmp[1] = "Following data would have been saved to the database: " + rowSummary.toString();
	                    results.rows().add(tmp);
                		
                	} else {
	                	wrk=PersistentWorkflowUtils.buildOpenWorkflow(user, item, ed);
	                	SaveItemHelper.unauthorizedSave(item,user, false, false,wrk.buildEvent());
	                	PersistentWorkflowUtils.complete(wrk, wrk.buildEvent());
	                	
	                	String[] tmp = new String[2];
	                    tmp[0] = "SUCCESS";
	                    tmp[1] = "Save Successful: " + rowSummary.toString();
	                    results.rows().add(tmp);
                	}
                } catch (Throwable e1) {
                    _log.error("",e1);
                    String[] tmp = new String[2];
                    tmp[0] = "ERROR";
                    tmp[1] = "Save Error: " + rowSummary.toString() + " " + e1.getMessage();
                    results.rows().add(tmp);
                    try {
                    	PersistentWorkflowUtils.fail(wrk, wrk.buildEvent());
                    } catch (Exception e){
                    	_log.error("Problem creating fail audit", e);
                    	tmp = new String[2];
                    	tmp[0] = "ERROR";
                    	tmp[1] = "Problem creating fail audit.";
                    	results.rows().add(tmp);
                    }
                }
            }

            //SUCEESS
            String[] tmp = new String[2];
        	tmp[0] = "Complete";
        	tmp[1] = "";
        	results.rows().add(tmp);
            
		    user.clearLocalCache();
		    
			try {
				MaterializedView.DeleteByUser(user);
			} catch (DBPoolException e) {
	            _log.error("",e);
	            _log.error("DBPoolException", e);
            	String[] tmp2 = new String[2];
            	tmp2[0] = "ERROR";
            	tmp2[1] = "DBPoolException: " + e.getMessage();
            	results.rows().add(tmp2);
			} catch (SQLException e) {
	            _log.error("",e);
	            String[] tmp2 = new String[2];
            	tmp2[0] = "ERROR";
            	tmp2[1] = "SQLException: " + e.getMessage();
            	results.rows().add(tmp2);
			} catch (Exception e) {
	            _log.error("",e);
	            String[] tmp2 = new String[2];
            	tmp2[0] = "ERROR";
            	tmp2[1] = "General Exception: " + e.getMessage();
            	results.rows().add(tmp2);
			}
			
        } catch (XFTInitException e) {
            _log.error("",e);
            String[] tmp = new String[2];
        	tmp[0] = "ERROR";
        	tmp[1] = "XFTInitException: " + e.getMessage();
        	results.rows().add(tmp);
        } catch (ElementNotFoundException e) {
            _log.error("",e);
            String[] tmp = new String[2];
        	tmp[0] = "ERROR";
        	tmp[1] = "ElementNotFoundException: " + e.getMessage();
        	results.rows().add(tmp);
        }
		return;
	}	
}
